from . import surfinBH
__author__ = surfinBH.__author__
__email__ = surfinBH.__email__
__copyright__ = surfinBH.__copyright__
__license__ = surfinBH.__license__
__version__ = surfinBH.__version__
__doc__ = surfinBH.__doc__

from ._loadFits import LoadFits
from ._loadFits import fits_collection
from ._loadFits import DownloadData
